%% agi.pl
%%
%% simple demonstration of prolog for AGI course
%% 
%% A simple family tree example -- the "hello world" of prolog.
%%
%% Example tested in Gnu Prolog, but should work on any standards-compliant
%% implementation.
%%
%% To play around, fire up prolog, and type
%%
%%      consult(agi).
%%
%% You may want to turn tracing on to understand what prolog is doing.
%%
%%      trace.
%%      notrace.  % turns tracing off


%% assertions of simple facts.
%% lowercase symbols are terminals.
male(bob).
male(ted).
male(john).
male(jack).

%% Note that prolog is a predicate logic. Statements are conditioned
%% on their arguments. "male" is a predicate, which may be true or
%% false depending on its argument. Because no non-terminals are
%% included, we are simply stating facts -- "the value of the
%% predicate male is true for arguments bob and ted".

female(carol).
female(alice).
female(jane).
female(jill).

%% similarly for females, we simply state some facts.

%% let's give them some kids
parent(bob, ted).
parent(carol, ted).
parent(bob, alice).
parent(carol, alice).
parent(john, jill).
parent(john, bob).
parent(jane, bob).
parent(alice, jack).

%% here's the family structure we have
%%           john    jane
%%           /   \   /
%%        jill    bob   carol
%%                  \   /
%%                ted,alice
%%                       \
%%                        jack

%% simple predicates
%% uppercase symbols are variables, but variables in the sense of mathematics
%% rather than normal programming variables. They are "bound" to terminals by
%% unification and then must keep the same value everywhere.

child(X, Y) :- parent(Y, X).
%% X is a child of Y iff Y is a parent of X.

father(X, Y) :- parent(X, Y), male(X).
%% X is the father of Y iff X is the parent of Y and X is a male

mother(X, Y) :- parent(X, Y), female(X).
%% X is the mother of Y iff X is the parent of Y and X is a female

son(X, Y) :- parent(Y, X), male(X).
%% note we've inverted the order of arguments to parent here

daughter(X, Y) :- parent(Y, X), female(X).
%% also here


%% the above predicates are simple, just relating one variable to another
%% directly. But we can also introduce new variables and let prolog search
%% for suitable bindings.

%% for example, consider grandparents. X is the grandparent of Y iff there
%% exists a third person Z such that X is the parent of Z and Z is the parent
%% of Y. 
grandparent(X, Y) :- parent(X, Z), parent(Z, Y).
grandfather(X, Y) :- grandparent(X, Y), male(X).
grandmother(X, Y) :- grandparent(X, Y), female(Y).
grandchild(X, Y) :- grandparent(Y, X).
grandson(X, Y) :- grandchild(X, Y), male(X).
granddaugher(X, Y) :- grandchild(X, Y), female(X).


%% recursive functors in prolog, we do all looping via recursion. To
%% walk up and down the family tree, we can do something like the
%% following.
ancestor(X, Y) :- parent(X, Y).
ancestor(X, Y) :- parent(X, Z), ancestor(Z, Y).
%% X is an ancestor of Y iff X is a parent of Y or a parent of some
%% ancestor of Y.
%%
%% Note how prolog does conditionals. Everything is done via pattern
%% matching, from top to bottom, left to right. When a match is found,
%% the right hand side is evaluated. If no binding can be found to
%% satisfy it, prolog backtracks and tries subsequent matching rules.

descendant(X, Y) :- child(X, Y).
descendant(X, Y) :- child(X, Z), descendant(Z, Y).
%% X is a descendant of Y iff X is a child of Y or the child of one
%% of Y's descendants.
%% Note we could have also written this as
%% descendant(X, Y) :- ancestor(Y, X).



%% 
%% search control
%% 

%% let's define some marriages
married(bob, carol).
married(john, jill).

%% there is a problem now...marriage should be symmetric. What happens
%% if we try "married(carol, bob)."?
%%
%% We can try to fix this by defining marriage as symmetric directly.
married(X, Y) :- married(Y, X).
%% what happens when you try a few examples with this new predicate?

%% prolog provides the cut operator (!) to allow manual control over
%% the search process. Sometimes it's necessary, either for correctness
%% or for efficiency, but its use should be minimized. Doesn't always
%% play well with the logical programming model.


%% computation
%%
%% Note that prolog is a fully general-purpose Turing-complete
%% programming language. We can write more substantial computations.
%% 
my_reverse([], []).
my_reverse([X|Xs], Ys) :- reverse(Xs, RXs), append(RXs, [X], Ys). 

%% this says that (a) the reverse of an empty list is an empty list, and
%% (b) the reverse of a list that starts with an X is a list Ys iff
%% there is a list RXs that is the reverse of the rest of the input
%% list that when appended to the first element of the list forms Ys.

%% For our purposes, we won't worry about writing significant computations
%% in prolog. We want to focus just on the logical inference engine.




%% Limitations of Prolog's inference engine
%%
%% Prolog is a deductive inference engine. Will only make an inference
%% if it can prove the inference is true.

%% Let's consider a predicate "knows", where knows(X,Y) should be true
%% if person X knows person Y. We could of course enumerate all the cases...
%% 
%% knows(bob,carol).
%% knows(carol, bob).
%% knows(john, jane).
%% knows(john, jill).
%% knows(jill, john).
%% ...
%%
%% this quickly gets tiring. Can we define a relation that holds more
%% generally?
%%
knows(X, Y) :- married(X, Y).		%% seems reasonable, I guess
knows(X, Y) :- parent(X, Y).
knows(X, Y) :- child(X, Y).
%% hmmm...there are certainly parents who don't know their children
%% and vice-versa.

%% the problem we have here is that deduction isn't really sufficient.
%% We want something that can express shades of gray. The system should
%% have or be able to learn things like "there is a high likelihood that
%% a parent knows his or her child".
%%
%% there are other statements that an intelligent reasoning agent should
%% be able to deal with. What is the likelihood that two people, X and Y,
%% are related? We as humans could assign some base probability here,
%% perhaps based on the size of the community, etc. Now assume that X
%% and Y know each other. Does this affect our estimate of the probability
%% that they are related? Yes! Perhaps only slightly, but if we believe
%% that relatives are more likely to know one another, then evidence of
%% knowing one another lends support to the proposition that they're
%% related.
%%
%% We could build this sort of system in prolog. It would be a lot of work,
%% but it could be done.
%%
%% First, we need non-binary truth values that can represent something like
%% probability or likelihood of truth.
%%
%% Next, we need inference rules that govern the types of conclusions the
%% system is able to draw from given knowledge.
%%
%% We also need rules governing how to assign truth values to these new
%% inferences we are able to draw.
%%
%% As a first-level approximation, NARS is a system for providing just
%% such a system. 
